from setuptools import setup, find_packages


VERSION = '0.0.1'
DESCRIPTION = 'Neural Network package'
LONG_DESCRIPTION = 'A package that allows to create a neural network, and train and predict with it. ' \
                   'It was developed in a university project and mainly used for gesture detection'

# Setting up
setup(
    name="netneural",
    version=VERSION,
    author="fachter (Felix)",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=["numpy", "pandas", "pyyaml", "matplotlib", "jupyter", "tqdm", "python-mnist", "seaborn"],
    keywords=['python', 'neural network', 'deep learning', 'nn', 'machine learning', 'neural net'],
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)
